# -*- coding: utf-8 -*-

import bz2
import csv

class OutputWriter(object):

    def __init__(self, metapopulation, filename, delimiter=',', compress=False):
        self.metapopulation = metapopulation
        self.filename = filename
        self.compress = compress

        if compress:
            self.filename += ".bz2"

            try:
                self.outfile = bz2.open(filename=self.filename, mode='wt',
                                        newline='', encoding='UTF-8')
            except AttributeError:
                self.outfile = bz2.BZ2File(self.filename, 'w')

        else:
            self.outfile = open(self.filename, mode='wt')

        self.writer = csv.writer(self.outfile, delimiter=delimiter)

    def update(self, time):
        pass

    def close(self):
        self.outfile.close()




import numpy as np

def bitstring_as_base10(a):
    return np.sum(a * 2**np.arange(len(a))[::-1])

def base10_as_bitarray(a):
    return np.array(list(map(int, bin(a)[2:])))

def hamming_distance(a, b):
    return bin(a ^ b).count('1')

def is_producer(g, bits):
    return g & 2**bits == 2**bits




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import datetime
import getpass
import os
import shutil
import signal
import sys
from time import time
from uuid import uuid4
from warnings import warn

from configobj import ConfigObj, ConfigObjError, flatten_errors
from validate import Validator

import networkx as nx
import numpy as np

import hankshaw

from hankshaw.Metapopulation import Metapopulation
from hankshaw.misc import write_run_information, append_run_information


def parse_arguments():
    """Parse command line arguments"""

    parser = argparse.ArgumentParser(prog='hankshaw.py',
                                     description='Run a simluation')
    parser.add_argument('--config', '-c', metavar='FILE', help='Configuration '\
                        'file to use', dest='configfile', default=None)
    parser.add_argument('--checkconfig', '-C', action='store_true',
                        default=False,
                        help='Check the given configuration file and quit (note: includes parameters specified with --param)')
    parser.add_argument('--genconfig', '-G', metavar='FILE',
                        help='Generate a configuration file with default values and quit (note: includes parameters specified with --param)')
    parser.add_argument('--data_dir', '-d', metavar='DIR',
                        help='Directory to store data (default: data)')
    parser.add_argument('--param', '-p', nargs=3, metavar=('SECTION', 'NAME',
                                                           'VALUE'),
                        action='append', help='Set a parameter value')
    parser.add_argument('--seed', '-s', metavar='S', help='Set the '\
                        'pseudorandom number generator seed', type=int)
    parser.add_argument('--quiet', '-q', action='store_true', default=False,
                       help='Suppress output messages')
    parser.add_argument('--version', action='version', version=hankshaw.__version__)

    args = parser.parse_args()

    return args


def main():
    start_time = time()

    cfgspec = os.path.join(hankshaw.__path__[0], 'configspec-v1.ini')

    # Get the command line arguments
    args = parse_arguments()

    if args.genconfig:
        if os.path.exists(args.genconfig):
            print("Error: File '{f}' already exists.".format(f=args.genconfig))
            sys.exit(1)

        config = ConfigObj(infile=args.genconfig, create_empty=True,
                           configspec=cfgspec)
        config.validate(Validator(), copy=True)
        config.write()
        print("Created configuration file '{f}'".format(f=args.genconfig))
        sys.exit(0)

    # Read the configuration file
    try:
        require_config = args.configfile is not None or args.checkconfig
        config = ConfigObj(infile=args.configfile,
                           configspec=cfgspec,
                           file_error=require_config)
    except (ConfigObjError, OSError) as e:
        print("Error: {e}".format(e=e))
        sys.exit(1)

    # Validate the configuration (this time, just copy in any missing values)
    validation = config.validate(Validator(), copy=True)


    # Add any parameters specified on the command line to the configuration
    if args.param:
        for p in args.param:
            config[p[0]][p[1]] = p[2]

    # Validate the modified configuration
    validation = config.validate(Validator(), copy=True)

    if validation != True:
        errors = flatten_errors(config, validation)
        print("Found {n} error(s) in configuration:".format(n=len(errors)))
        for (section_list, key, _) in errors:
            if key is not None:
                print("\t* Invalid value for '{k}' in Section '{s}'".format(k=key, s=section_list[0]))
            else:
                print("\t* Missing required section '{s}'".format(s=section_list[0]))

        sys.exit(2)

    if args.checkconfig:
        print("No errors found in configuration file '{f}'".format(f=args.configfile))
        sys.exit(0)

    # If the random number generator seed specified, add it to the config,
    # overwriting any previous value. Otherwise, if it wasn't in the
    # supplied configuration file, create one.
    if args.seed:
        config['Simulation']['seed'] = args.seed
    elif config['Simulation']['seed'] == 0:
        seed = np.random.randint(low=0, high=np.iinfo(np.uint32).max)
        config['Simulation']['seed'] = seed

    np.random.seed(seed=config['Simulation']['seed'])

    # Generate a universally unique identifier (UUID) for this run
    config['Simulation']['UUID'] = str(uuid4())


    # If the data directory is specified, add it to the config, overwriting any
    # previous value
    if args.data_dir:
        config['Simulation']['data_dir'] = args.data_dir


    # If the data_dir already exists, append the current date and time to
    # data_dir, and use that. Afterwards, create the directory.
    if os.path.exists(config['Simulation']['data_dir']):
        newname = '{o}-{d}'.format(o=config['Simulation']['data_dir'],
                                   d=datetime.datetime.now().strftime("%Y%m%d%H%M%S"))
        msg = '{d} already exists. Using {new} instead.'.format(d=config['Simulation']['data_dir'],
                                                                new=newname)
        warn(msg)
        config['Simulation']['data_dir'] = newname

    os.mkdir(config['Simulation']['data_dir'])


    # Write the configuration file
    config.filename = os.path.join(config['Simulation']['data_dir'], 'run.cfg')
    config.write()


    # Write information about the run
    infofile = os.path.join(config['Simulation']['data_dir'], 'run_info.txt')
    write_run_information(filename=infofile, config=config)


    # Create and initialize the metapopulation
    m = Metapopulation(config=config)


    # Print a status message when SIGINFO (ctrl-T) is received on BSD or
    # OS X systems or when SIGUSR1 is received on POSIX systems
    def handle_siginfo(signum, frame):
        print("Cycle {c}: Size {ps}, {pc:.0%} cooperators".format(c=m.time, ps=m.size(), pc=m.prop_producers()))

    signal.signal(signal.SIGUSR1, handle_siginfo)
    if hasattr(signal, 'SIGINFO'):
        signal.signal(signal.SIGINFO, handle_siginfo)


    # Run the simulation
    for t in range(config['Simulation']['num_cycles']):
        m.cycle()

        if not args.quiet:
            msg = "[{t}] {m}".format(t=t, m=m)
            print(msg)

        if config['Simulation']['stop_when_empty'] and m.size() == 0:
            break

    m.write_logfiles()
    m.cleanup()

    rt_string = 'Run Time: {t}\n'.format(t=datetime.timedelta(seconds=time()-start_time))
    append_run_information(filename=infofile, string=rt_string)


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-

from hankshaw.OutputWriter import OutputWriter


class EnvChangeOutput(OutputWriter):

    def __init__(self, metapopulation, filename='environmental_change.csv',
                 delimiter=',', compress=False):
        super(EnvChangeOutput, self).__init__(metapopulation=metapopulation,
                                              filename=filename,
                                              delimiter=delimiter,
                                              compress=compress)

        self.writer.writerow(['Time', 'EnvironmentChanged'])

    def update(self, time):
        self.writer.writerow([time, self.metapopulation.environment_changed])




# -*- coding: utf-8 -*-

from hankshaw.OutputWriter import OutputWriter


class MetapopulationOutput(OutputWriter):

    def __init__(self, metapopulation, filename='metapopulation.csv',
                 delimiter=',', compress=False):
        super(MetapopulationOutput, self).__init__(metapopulation=metapopulation,
                                                   filename=filename,
                                                   delimiter=delimiter,
                                                   compress=compress)

        self.writer.writerow(['Time', 'Births', 'Size', 'CooperatorProportion',
                              'MaxCooperatorFitness', 'MaxDefectorFitness'])

    def update(self, time):
        fits = self.metapopulation.max_fitnesses()
        self.writer.writerow([time, self.metapopulation.num_births,
                              self.metapopulation.size(),
                              self.metapopulation.prop_producers(),
                              max(fits[0]), max(fits[1])])




# -*- coding: utf-8 -*-

"""Miscellaneous functions"""

import datetime
import getpass
import platform
import sys

import hankshaw
import networkx as nx
import numpy as np



def write_run_information(filename, config):
    """ Write information about the run and software environment """

    with open(filename, 'w') as infofile:
        infostr = 'Generated by {whom} at {when}\n'
        infofile.write(infostr.format(when=datetime.datetime.now().isoformat(),
                                        whom=getpass.getuser()))
        infofile.write('Host: {p}\n'.format(p=platform.node()))
        infofile.write('Platform: {p}\n'.format(p=platform.platform()))
        infofile.write('Hankshaw Model Version: {v}\n'.format(v=hankshaw.__version__))
        infofile.write('Python Version: {v}\n'.format(v=".".join([str(n) for n in sys.version_info[:3]])))
        infofile.write('NumPy Version: {v}\n'.format(v=np.version.version))
        infofile.write('NetworkX Version: {v}\n'.format(v=nx.__version__))
        infofile.write('Command: {cmd}\n'.format(cmd=' '.join(sys.argv)))
        infofile.write('Random Seed: {seed}\n'.format(seed=config['Simulation']['seed']))
        infofile.write('UUID: {uuid}\n'.format(uuid=config['Simulation']['UUID']))

def append_run_information(filename, string):
    """ Add additional information about the run and software environment """

    with open(filename, 'a') as infofile:
        infofile.write(string)




# -*- coding: utf-8 -*-

from hankshaw.OutputWriter import OutputWriter


class FitnessOutput(OutputWriter):

    def __init__(self, metapopulation, filename='max_fitness.csv',
                 delimiter=',', compress=False):
        super(FitnessOutput, self).__init__(metapopulation=metapopulation,
                                            filename=filename,
                                            delimiter=delimiter,
                                            compress=compress)

        self.writer.writerow(['Time', 'Producers', 'Nonproducers'])

    def update(self, time):
        maxfit = self.metapopulation.max_fitnesses()
        self.writer.writerow([time, max(maxfit[0]), max(maxfit[1])])



# -*- coding: utf-8 -*-

import numpy as np
from numpy import sum as nsum
from numpy import zeros as zeros
from numpy.random import binomial
from numpy.random import multinomial

from hankshaw import genome


class Population(object):
    """Represent a population within a metapopulation

    A population is a collection of individuals. Each individual is represented
    by a number. The binary representation of that number defines that
    individual's genotype. The state of the highest order bit determines whether
    (1) or not (0) that individual is a producer.

    * genome_length: the length of the genome. The production allele is added to
        this, so the number of genotypes is 2^(genome_length+1)
    * stress_survival_rate: the probability of an individual surviving a change
        of environment (stress)
    * mutation_rate_social: the probability of a mutation (bit flip) occuring at
        the social locus
    * mutation_rate_adaptation: the probability of a mutation (bit flip) at a
        non-social locus
    * capacity_min: the minimum size of a fully-grown population. This occurs
        when there are no producers
    * capacity_max: the maximum size of a fully-grown population. This occurs
        when a population consists entirely of producers
    * production_cost: the fitness cost of production. This manifests itself as
        a decrease in growth rate
    * initialize: How to initialize the population
        empty: the population will have no individuals


    """

    def __init__(self, metapopulation, config):
        """Initialize a Population object"""
        self.metapopulation = metapopulation
        self.config = config

        self.genome_length = config['Population']['genome_length']
        self.stress_survival_rate = config['Population']['stress_survival_rate']
        self.mutation_rate_social = config['Population']['mutation_rate_social']
        self.mutation_rate_adaptation = config['Population']['mutation_rate_adaptation']
        self.dilution_factor = config['Population']['dilution_factor']
        self.dilution_prob_min = config['Population']['dilution_prob_min']
        self.capacity_min = config['Population']['capacity_min']
        self.capacity_max = config['Population']['capacity_max']
        self.capacity_shape = config['Population']['capacity_shape']
        self.production_cost = config['Population']['production_cost']
        self.initialize = config['Population']['initialize']

        # Create an empty population
        if self.initialize.lower() == 'empty':
            self.empty()
        elif self.initialize.lower() == 'random':
            self.randomize()

        self.delta = zeros(self.abundances.size, dtype=np.int32)
        self.diluted = True


    def __repr__(self):
        """Return a string representation of a Population object"""
        res = "Population: Size {s}, {p:.1%} producers".format(s=self.size(),
                                                               p=self.prop_producers())
        return res


    def empty(self):
        """Empty a population"""
        self.abundances = zeros(2**(self.genome_length + 1), dtype=np.uint32)


    def randomize(self):
        """Create a random population"""
        self.abundances = np.random.random_integers(low=0,
                                                    high=self.capacity_min,
                                                    size=2**(self.genome_length+1))


    def dilute(self):
        """Dilute a population
        
        dilute dilutes the population by the dilution factor, which is specified
        in the Population section of the configuration as dilution_factor.

        """
        if self.is_empty():
            return

        self.diluted = False
        prob_dilute = 1

        if self.dilution_prob_min < 1:
            prob_dilute = self.dilution_prob_min + (1.0 - self.dilution_prob_min) * self.prop_producers()

        if self.dilution_prob_min == 1 or binomial(n=1, p=prob_dilute, size=1)[0]:
            self.abundances = binomial(self.abundances, self.dilution_factor)
            self.diluted = True


    def grow(self):
        """Grow the population to carrying capacity
        
        The final population size is determined based on the proportion of
        producers present. This population is determined by drawing from a
        multinomial with the probability of each genotype proportional to its
        abundance times its fitness.
        """

        if self.is_empty() or not self.diluted:
            return

        landscape = self.metapopulation.fitness_landscape

        final_size = self.capacity_min + \
                (self.capacity_max - self.capacity_min) * \
                (self.prop_producers()**self.capacity_shape)

        grow_probs = self.abundances * (landscape/nsum(landscape))

        if nsum(grow_probs) > 0:
            norm_grow_probs = grow_probs/nsum(grow_probs)
            self.abundances = multinomial(final_size, norm_grow_probs, 1)[0]

        self.metapopulation.num_births += self.size()


    def mutate(self):
        """Mutate a Population
        
        Each genotype mutates to another with probability inversely proportional
        to the Hamming distance (# different bits in binary representation)
        between them. The distances between all pairs of genotypes is
        pre-calculated at the beginning of a run and stored in
        metapopulation.mutation_probs.
        
        """

        if self.is_empty():
            return

        if not self.diluted:
            return

        mutated_population = zeros(self.abundances.size, dtype=np.uint32)

        for i in np.nonzero(self.abundances)[0]:
            mutated_population = np.add(mutated_population,
                                        multinomial(self.abundances[i],
                                                    self.metapopulation.mutation_probs[i],
                                                    size=1)[0])

        self.abundances = mutated_population


    def select_migrants(self, migration_rate):
        """Select individuals to migrate
                                    
        Select genotypes to migrate. The amount of each genotype that migrates
        is chosen in proportion to that genotype's abundance.
                                    
        """

        assert migration_rate >= 0 and migration_rate <= 1

        return binomial(self.abundances, migration_rate)


    def remove_emigrants(self, emigrants):
        """Remove emigrants from the population
                                    
        remove_emigrants removes the given emigrants from the population. The
        genotypes are not immediately removed to the population, but their
        counts are placed in a temporary area until census() is called.

        """
        self.delta -= emigrants


    def add_immigrants(self, immigrants):
        """Add immigrants to the population
                                    
        add_immigrants adds the given immigrants to the population. The new
        genotypes are not immediately added to the population, but placed in
        a temporary area until census() is called.

        """
        self.delta += immigrants


    def census(self):
        """Update the population's abundances after migration
        
        When migration occurs, the immigrants and emigrants are not directly
        accounted for in the list of genotype abundances. This function adds
        immigrants and removes emigrants to/from the abundances.
                                    
        """

        self.abundances = np.add(self.abundances, self.delta)
        self.delta = zeros(self.abundances.size, dtype=np.int32)


    def reset_loci(self):
        """Reset the loci of the population to all zeros

        When an environment changes, the population is not yet adapted to it.
        This function captures this change by resetting all fitness-encoding
        loci to zero.
        """
        num_producers = self.num_producers()                                    
        num_nonproducers = self.num_nonproducers()                              

        self.abundances = zeros(self.abundances.size, dtype=np.int)      
        self.abundances[0] = num_nonproducers                                   
        self.abundances[2**self.genome_length] = num_producers              


    def bottleneck(self, survival_rate):
        """ Pass the population through a bottleneck

        This function passes the population through a bottleneck. The
        probability of survival is specified as the survival_rate parameter
        [0,1]. 
        """

        assert survival_rate >= 0
        assert survival_rate <= 1

        self.abundances = binomial(self.abundances, survival_rate)


    def size(self):
        """Get the size of the population"""
        return self.abundances.sum()


    def __len__(self):
        return self.abundances.sum()


    def is_empty(self):
        """Return whether or not the population is empty"""
        return self.abundances.sum() == 0


    def num_producers(self):
        """Get the number of producers"""
        return self.abundances[2**self.genome_length:].sum()


    def num_nonproducers(self):
        """Get the number of non-producers"""
        return self.abundances[:2**self.genome_length].sum()


    def prop_producers(self):
        """Get the proportion of producers"""
        popsize = self.abundances.sum()
        
        if popsize == 0:
            return 'NA'
        else:
            return 1.0 * self.num_producers() / popsize


    def average_fitness(self):
        """Get the average fitness in the population"""

        popsize = self.size()
        landscape = self.metapopulation.fitness_landscape

        if popsize == 0:
            return 'NA'
        else:
            return np.sum(self.abundances * landscape)/popsize


    def max_fitnesses(self):
        """Get the maximum fitness among producers and non-producers"""

        popsize = self.size()

        if popsize == 0:
            return (0,0)

        # Get the fitnesses of genotypes present in the population
        fitnesses = np.array(self.abundances > 0, dtype=int) * self.metapopulation.fitness_landscape

        max_producer = fitnesses[2**self.genome_length:].max()
        max_nonproducer = fitnesses[:2**self.genome_length].max()

        return (max_producer, max_nonproducer)




# -*- coding: utf-8 -*-

import networkx as nx


def moore_lattice(rows, columns, radius=1, periodic=False):
    """ Return the 2d lattice graph of rows x columns nodes, each connected to
    its nearest 8 neighbors within a given radius.  Optional argument
    periodic=True will connect boundary nodes via periodic boundary conditions.

    Parameters:

    *rows*
        The number of rows to be in the graph
    *columns*
        The number of columns to be in the graph
    *radius*
        The radius of interactions in the graph (there will be an edge
        between a node and all other nodes within N hops)
    *periodic*
        Prevent edge effects using periodic boundaries

    """
    G = nx.empty_graph()
    G.name = "Moore Lattice: {r} rows, {c} columns, radius={rx}".format(r=rows,
                                                                        c=columns,
                                                                        rx=radius)
    
    if periodic:
        G.name += ' with periodoc boundaries' 

    G.add_nodes_from(list(range(rows * columns)))

    for n in G.nodes():
        myrow = n // columns
        mycol = n % columns

        for r in range(myrow - radius, myrow + radius + 1):
            if periodic == False and (r < 0 or r >= rows):
                continue

            for c in range(mycol - radius, mycol + radius + 1):
                if periodic == False and (c < 0 or c >= columns):
                    continue

                neighbor = (columns * (r % rows)) + (c % columns)

                if n != neighbor:
                    G.add_edge(n, neighbor)

    return G


def vonneumann_lattice(rows, columns, periodic=False):
    """ Return the 2d lattice graph of rows x columns nodes, each connected to
    its nearest 4 neighbors.  Optional argument periodic=True will connect
    boundary nodes via periodic boundary conditions.

    Parameters:

    *rows*
        The number of rows to be in the graph
    *columns*
        The number of columns to be in the graph
    *periodic*
        Prevent edge effects using periodic boundaries
    """

    g = nx.grid_2d_graph(m=columns, n=rows, periodic=periodic)
    g = nx.convert_node_labels_to_integers(g)
    g.name = "VonNeumann Lattice: {r} rows, {c} columns".format(r=rows,
                                                                c=columns)

    if periodic:
        g.name += ' with periodic boundaries'

    return g


def smallworld(size, neighbors, edgeprob, seed=None):
    assert size > 0
    assert neighbors >= 0
    assert edgeprob >= 0 and edgeprob <= 1

    if seed:
        g = nx.newman_watts_strogatz_graph(n=size, k=neighbors, p=edgeprob,
                                           seed=seed)

    else:
        g = nx.newman_watts_strogatz_graph(n=size, k=neighbors, p=edgeprob)

    g.name = 'Small World network: {s} nodes, {n} neighbors, ' \
             '{p} edge probability'.format(s=size, n=neighbors, p=edgeprob)

    return g

def regular(size, degree, seed=None):
    assert size > 0
    assert degree >= 0

    if seed:
        g = nx.random_regular_graph(d=degree, n=size, seed=seed)
    else:
        g = nx.random_regular_graph(d=degree, n=size)

    g.name = 'Random Regular Graph: {n} nodes, {d} degree'.format(n=size,
                                                                  d=degree)
    return g



# -*- coding: utf-8 -*-

import numpy as np

from hankshaw.OutputWriter import OutputWriter
import hankshaw.genome


class GenotypesOutput(OutputWriter):

    def __init__(self, metapopulation, filename='genotypes.csv', delimiter=',',
                 compress=False):

        super(GenotypesOutput, self).__init__(metapopulation=metapopulation,
                                              filename=filename,
                                              delimiter=delimiter,
                                              compress=compress)

        self.genome_length = self.metapopulation.config['Population']['genome_length']

        self.writer.writerow(['Time', 'Genotype', 'AvgAbundance', 'IsProducer'])

    def update(self, time):
        abundances = []
        abundances = np.array([d['population'].abundances for n, d in self.metapopulation.topology.nodes_iter(data=True)])
        av = np.average(abundances, 0)
        
        for i in range(2**(self.genome_length+1)):
            isprod = genome.is_producer(i, self.genome_length)
            genotype = i & 2**(self.genome_length)-1
            self.writer.writerow([time, genotype, av[i], isprod])




# -*- coding: utf-8 -*-

from hankshaw.OutputWriter import OutputWriter


class PopulationOutput(OutputWriter):

    def __init__(self, metapopulation, filename='population.csv',
                 delimiter=',', compress=False):
        super(PopulationOutput, self).__init__(metapopulation=metapopulation,
                                               filename=filename,
                                               delimiter=delimiter,
                                               compress=compress)

        self.writer.writerow(['Time', 'Population', 'Size', 'Producers',
                              'PropProducers', 'NonProducers',
                              'PropNonProducers', 'AvgFitness'])

    def update(self, time):
        for n, d in self.metapopulation.topology.nodes_iter(data=True):
            size = len(d['population'])

            if size == 0:
                num_producers = 0
                num_nonproducers = 0
                prop_producers = 'NA'
                prop_nonproducers = 'NA'
                average_fitness = 'NA'
            else:
                num_producers = d['population'].num_producers()
                num_nonproducers = size - num_producers
                prop_producers = 1.0*num_producers/size
                prop_nonproducers = 1.0*num_nonproducers/size
                average_fitness = d['population'].average_fitness()

            self.writer.writerow([time, n, size, num_producers, prop_producers,
                                  num_nonproducers, prop_nonproducers,
                                  average_fitness])




# -*- coding: utf-8 -*-

VERSION = (1, 1, 1)
__version__ = ".".join(map(str, VERSION[0:3])) + "".join(VERSION[3:])
__license__ = "BSD"

from hankshaw.Metapopulation import *
from hankshaw.Population import *

from hankshaw.OutputWriter import *
from hankshaw.EnvChangeOutput import *
from hankshaw.FitnessOutput import *
from hankshaw.GenotypesOutput import *
from hankshaw.MetapopulationOutput import *
from hankshaw.PopulationOutput import *

import hankshaw.genome
import hankshaw.misc
import hankshaw.topology



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Script for installing the model.

To install, run:

    python setup.py install

"""

# Modified from https://github.com/pypa/sampleproject/blob/master/setup.py

from setuptools import setup, find_packages

from codecs import open
from os import path
import sys


if sys.argv[-1] == 'setup.py':
    print("To install pyfttt, run 'python setup.py install'\n")

if sys.version_info[:2] < (2, 7):
    print("pyfttt requires Python 2.7 or later (%d.%d detected)." % sys.version_info[:2])
    sys.exit(-1)


here = path.abspath(path.dirname(__file__))

setup(
    name='hankshaw',
    version='1.1.1',
    description='Model for The Evolution of Cooperation by the Hankshaw Effect',
    long_description='Simulation model described and used in The Evolution of Cooperation by the Hankshaw Effect. In each simulation, a metapopulation of populations grow and evolve.',
    url='https://github.com/briandconnelly/hankshaweffect',
    author='Brian D. Connelly',
    author_email='bdc@bconnelly.net',
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Life',
        'License :: OSI Approved :: BSD License'
    ],

    # What does your project relate to?
    keywords='evolution simulation metapopulation',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    install_requires=['numpy>=1.9.1', 'networkx>=1.9.1', 'configobj>=5.0.6'],

    extras_require={},
    package_data={},
    data_files=[],

    include_package_data=True,

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'hankshaw=hankshaw.cmd_script:main',
        ],
    },
)




# -*- coding: utf-8 -*-

import os

import networkx as nx
import numpy as np
from numpy.random import binomial, choice as nchoice, random_integers

from hankshaw import genome
from hankshaw.Population import Population
import hankshaw.topology as topology
from hankshaw.PopulationOutput import PopulationOutput
from hankshaw.MetapopulationOutput import MetapopulationOutput
from hankshaw.GenotypesOutput import GenotypesOutput
from hankshaw.FitnessOutput import FitnessOutput
from hankshaw.EnvChangeOutput import EnvChangeOutput


class Metapopulation(object):

    def __init__(self, config):
        """Initialize a Metapopulation object"""
        self.config = config
        self.time = 0
        self.num_births = 0

        self.migration_rate = self.config['Metapopulation']['migration_rate']
        self.migration_dest = self.config['Metapopulation']['migration_dest']
        self.migration_p_far = self.config['Metapopulation']['migration_p_far']
        self.topology_type = self.config['Metapopulation']['topology']

        if self.topology_type.lower() == 'moore':
            width = self.config['MooreTopology']['width']
            height = self.config['MooreTopology']['height']
            periodic = self.config['MooreTopology']['periodic']
            radius = self.config['MooreTopology']['radius']

            self.topology = topology.moore_lattice(rows=height, columns=width,
                                                   radius=radius,
                                                   periodic=periodic)

        elif self.topology_type.lower() == 'vonneumann':
            width = self.config['VonNeumannTopology']['width']
            height = self.config['VonNeumannTopology']['height']
            periodic = self.config['VonNeumannTopology']['periodic']

            self.topology = topology.vonneumann_lattice(rows=height,
                                                        columns=width,
                                                        periodic=periodic)

        elif self.topology_type.lower() == 'smallworld':
            size = self.config['SmallWorldTopology']['size']
            neighbors = self.config['SmallWorldTopology']['neighbors']
            edgeprob = self.config['SmallWorldTopology']['edgeprob']

            if self.config['SmallWorldTopology']['seed'] == 0:
                self.config['SmallWorldTopology']['seed'] = self.config['Simulation']['seed']

            seed = self.config['SmallWorldTopology']['seed']

            self.topology = topology.smallworld(size=size, neighbors=neighbors,
                                                edgeprob=edgeprob, seed=seed)


        elif self.topology_type.lower() == 'complete':
            self.topology = nx.complete_graph(n=self.config['CompleteTopology']['size'])


        elif self.topology_type.lower() == 'regular':
            size = self.config['RegularTopology']['size']
            degree = self.config['RegularTopology']['degree']

            if self.config['RegularTopology']['seed'] == 0:
                self.config['RegularTopology']['seed'] = self.config['Simulation']['seed']

            seed = self.config['RegularTopology']['seed']

            self.topology = topology.regular(size=size, degree=degree,
                                             seed=seed)


        # Export the structure of the topology, allowing the topology to be
        # re-created. This is especially useful for randomly-generated
        # topologies.
        if self.config['Metapopulation']['export_topology']:
            nx.write_gml(self.topology, os.path.join(self.config['Simulation']['data_dir'], 'topology.gml'))


        # Store the probabilities of mutations between all pairs of genotypes
        self.mutation_probs = self.get_mutation_probabilities()

        # Create the fitness landscape
        self.fitness_landscape = self.build_fitness_landscape()

        initial_state = self.config['Metapopulation']['initial_state']
        genome_length = self.config['Population']['genome_length']
        max_cap = self.config['Population']['capacity_max']
        min_cap = self.config['Population']['capacity_min']
        initial_producer_proportion = self.config['Population']['initial_producer_proportion']


        # Create each of the populations
        for n, d in self.topology.nodes_iter(data=True):
            d['population'] = Population(metapopulation=self, config=config)

            if initial_state == 'corners':
                # Place all producers in one corner and all non-producers in
                # the other
                if n == 0:
                    d['population'].abundances[2**genome_length] = max_cap
                    d['population'].dilute()
                elif n == len(self.topology)-1:
                    d['population'].abundances[0] = min_cap
                    d['population'].dilute()

            elif initial_state == 'stress':
                cap = int(min_cap + ( (max_cap - min_cap) * initial_producer_proportion))
                num_producers = int(cap * initial_producer_proportion)
                num_nonproducers = cap - num_producers

                d['population'].abundances[0] = num_nonproducers
                d['population'].abundances[2**genome_length] = num_producers
                d['population'].bottleneck(survival_rate=self.config['Population']['stress_survival_rate'])

        # How frequently should the metapopulation be mixed?
        self.mix_frequency = self.config['Metapopulation']['mix_frequency']


        # How frequently should the environment be changed?
        self.environment_changed = False
        self.env_change_frequency = self.config['Metapopulation']['env_change_frequency']

        if self.env_change_frequency > 0:
            self.next_env_change_cycle = np.round(np.random.exponential(scale=self.env_change_frequency, size=1)[0]).astype(int)


        data_dir = self.config['Simulation']['data_dir']

        # log_objects is a list of any logging objects used by this simulation
        self.log_objects = []

        if self.config['MetapopulationLog']['enabled']:
            fname = self.config['MetapopulationLog']['filename']
            freq = self.config['MetapopulationLog']['frequency']
            compress = self.config['MetapopulationLog']['compress']
            self.log_objects.append((freq, MetapopulationOutput(metapopulation=self,
                                                                filename=os.path.join(data_dir, fname),
                                                                compress=compress)))

        if self.config['PopulationLog']['enabled']:
            fname = self.config['PopulationLog']['filename']
            freq = self.config['PopulationLog']['frequency']
            compress = self.config['PopulationLog']['compress']
            self.log_objects.append((freq, PopulationOutput(metapopulation=self,
                                                            filename=os.path.join(data_dir, fname),
                                                            compress=compress)))

        if self.config['GenotypeLog']['enabled']:
            fname = self.config['GenotypeLog']['filename']
            freq = self.config['GenotypeLog']['frequency']
            compress = self.config['GenotypeLog']['compress']
            self.log_objects.append((freq, GenotypesOutput(metapopulation=self,
                                                           filename=os.path.join(data_dir, fname),
                                                           compress=compress)))

        if self.config['FitnessLog']['enabled'] :
            fname = self.config['FitnessLog']['filename']
            freq = self.config['FitnessLog']['frequency']
            compress = self.config['FitnessLog']['compress']
            self.log_objects.append((freq, FitnessOutput(metapopulation=self,
                                                         filename=os.path.join(data_dir, fname),
                                                           compress=compress)))

        if self.config['EnvChangeLog']['enabled']:
            fname = self.config['EnvChangeLog']['filename']
            freq = self.config['EnvChangeLog']['frequency']
            compress = self.config['EnvChangeLog']['compress']
            self.log_objects.append((freq, EnvChangeOutput(metapopulation=self,
                                                           filename=os.path.join(data_dir, fname),
                                                           compress=compress)))


    def __repr__(self):
        """Return a string representation of the Metapopulation object"""
        prop_producers = self.prop_producers()

        if prop_producers == 'NA':
            res = "Metapopulation: Size {s}, NA% producers".format(s=self.size())
        else:
            maxfit = self.max_fitnesses()
            maxfit_p = max(maxfit[0]) / max(self.fitness_landscape)
            maxfit_np = max(maxfit[1]) / max(self.fitness_landscape)
            #res = "Metapopulation: Size {s}, {p:.1%} producers".format(s=self.size(),
            #                                                        p=self.prop_producers())

            if maxfit_p > maxfit_np:
                symbol = '>'
            elif maxfit_p < maxfit_np:
                symbol = '<'
            else:
                symbol = '='

            res = "Metapopulation: Size {s}, {p:.1%} producers. w(P): {mp:.2} "\
                  "{sym} w(Np): {mnp:.2}.".format(s=self.size(), p=self.prop_producers(),
                                                 mp=maxfit_p, mnp=maxfit_np, sym=symbol)

        return res


    def build_fitness_landscape(self):
        """Build a fitness landscape

        """

        genome_length = self.config['Population']['genome_length']
        base_fitness = self.config['Population']['base_fitness']
        production_cost = self.config['Population']['production_cost']
        benefit_nonzero = self.config['Population']['benefit_nonzero']

        effects = np.append(-1.0*production_cost,
                            np.repeat(benefit_nonzero, genome_length))

        landscape = np.zeros(2**(genome_length + 1))

        for i in range(2**(genome_length + 1)):
            genotype = genome.base10_as_bitarray(i)
            genotype = np.append(np.zeros(len(effects) - len(genotype)),
                                 genotype)
            landscape[i] = sum(genotype * effects) + (base_fitness + production_cost)

        return landscape


    def get_mutation_probabilities(self):
        """Get a table of probabilities among all pairs of genotypes
        
        This works by first generating the Hamming distances between all of the
        possible genotypes. These distances are then used to calculate the
        probabilities of mutating by:

            (1-mu)^(#matching bits) * mu^(#different bits)

        Where #matching bits is the genome length - Hamming distance and
        #different bits is the Hamming distance.
        
        """

        genome_length = self.config['Population']['genome_length']
        mutation_rate_social = self.config['Population']['mutation_rate_social']
        mutation_rate_adaptation = self.config['Population']['mutation_rate_adaptation']

        S = np.vstack((np.array([[0]*2**genome_length + [1]*2**genome_length]).repeat(repeats=2**genome_length, axis=0),
                       np.array([[1]*2**genome_length + [0]*2**genome_length]).repeat(repeats=2**genome_length, axis=0)))


        # TODO: to handle things like P->NP but not NP->P (or vice versa), just
        # manipulate the mr vector.

        # Get the pairwise Hamming distance for all genotypes
        hamming_v = np.vectorize(genome.hamming_distance)
        genotypes = np.arange(start=0, stop=2**(genome_length+1))
        xx, yy = np.meshgrid(genotypes, genotypes)
        hamming_distances = hamming_v(xx, yy)

        # nonsocial_hd is a matrix containing the pairwise Hamming distances
        # between all genomes considering only the non-social loci
        nonsocial_hd = hamming_distances - S

        # mr is a matrix where each element contains the probability of mutating
        # from one genome to the other.
        # mr = npower(1-m1, L-NS) * npower(m1, NS) * npower(1-m2, S2) * npower(m2, S)

        npower = np.power
        mr = npower(1-mutation_rate_adaptation, genome_length-nonsocial_hd) *\
                npower(mutation_rate_adaptation, nonsocial_hd) *\
                npower(1-mutation_rate_social, S==0) *\
                npower(mutation_rate_social, S)

        return mr


    def dilute(self):
        """Dilute the metapopulation

        Dilute the metapopulation by diluting each population by the dilution
        factor specified with the dilution_factor option in the Population
        section of the configuration file.

        """
        for n, d in self.topology.nodes_iter(data=True):
            d['population'].dilute()


    def mix(self):
        """Mix the population

        Mix the population. The abundances at all populations are combined and
        re-distributed.
        """

        abundances = np.zeros(self.fitness_landscape.size, dtype=np.int)

        for n, d in self.topology.nodes_iter(data=True):
            abundances += d['population'].abundances

        for n, d in self.topology.nodes_iter(data=True):
            d['population'].abundances = binomial(abundances, 1.0/len(self.topology))


    def grow(self):
        """Grow the metapopulation ...."""
        for n, d in self.topology.nodes_iter(data=True):
            d['population'].grow()


    def mutate(self):
        """Mutate the metapopulation ...."""
        for n, d in self.topology.nodes_iter(data=True):
            d['population'].mutate()


    def migrate(self, single_dest=True):
        """Migrate individuals among the populations
        
        * single_dest: if True (default), all migrants will go to a single
            neighbor population. Otherwise, migrants will be distributed among
            all neighbors. 
        
        """
        if self.migration_rate == 0 or self.topology.number_of_nodes() < 2:
            return

        for n, d in self.topology.nodes_iter(data=True):
            pop = d['population']

            if self.topology.degree(n) == 0:
                return

            # Migrate everything to one neighboring population
            if self.migration_dest.lower() == 'single':
                migrants = pop.select_migrants(migration_rate=self.migration_rate)

                if binomial(n=1, p=self.migration_p_far, size=None) == 1:
                    neighbor_index = random_integers(low=0,high=self.topology.number_of_nodes()-1)
                else:
                    neighbor_index = nchoice(self.topology.neighbors(n))

                neighbor = self.topology.node[neighbor_index]['population']
                neighbor.add_immigrants(migrants)
                pop.remove_emigrants(migrants)

            # Distribute the migrants among the neighboring populations
            elif self.migration_dest.lower() == 'neighbors':
                num_neighbors = self.topology.degree(n)
                for neighbor_node in self.topology.neighbors_iter(n):
                    if binomial(n=1, p=self.migration_p_far, size=None) == 1: 
                        neighbor_node = random_integers(low=0,high=self.topology.number_of_nodes()-1)
                    migrants = pop.select_migrants(migration_rate=self.migration_rate/num_neighbors)
                    neighbor = self.topology.node[neighbor_node]['population']
                    neighbor.add_immigrants(migrants)
                    pop.remove_emigrants(migrants)


    def census(self):
        """Update each population's abundance to account for migration"""
        for n, d in self.topology.nodes_iter(data=True):
            d['population'].census()


    def cycle(self):
        """Cycle the metapopulation

        In each cycle, the metapopulation cycles its state by diluting each
        population, allowing each population to grow to capacity, mutate each
        population, and then migrating among populations.

        """
        self.write_logfiles()

        self.grow()
        self.mutate()
        self.migrate()
        self.census()

        if self.mix_frequency > 0 and self.time > 0 and \
                (self.time % self.mix_frequency == 0):
            self.mix()

        if self.env_change_frequency > 0 and self.time == self.next_env_change_cycle:
            self.change_environment()
            self.environment_changed = True
            self.next_env_change_cycle = self.time + np.round(np.random.exponential(scale=self.env_change_frequency, size=1)[0]).astype(int)
        else:
            self.dilute()
            self.environment_changed = False

        self.time += 1


    def change_environment(self):
        """Change the environment

        The change_environment function changes the environment for the
        metapopulation. This re-generates the fitness landscape and zeros out
        all fitness-encoding loci. This is meant to represent the metapopulation
        being subjected to different selective pressures. The number of
        individuals of each genotype that survive this event are proportional to
        the abundance of that genotype times the mutation rate (representing
        individuals that acquired the mutation that allows them to persist).
        """

        self.fitness_landscape = self.build_fitness_landscape()

        for n, d in self.topology.nodes_iter(data=True):
            d['population'].bottleneck(survival_rate=self.config['Population']['stress_survival_rate'])
            d['population'].reset_loci()


    def size(self):
        """Return the size of the metapopulation

        The size of the metapopulation is the sum of the sizes of the
        subpopulations
        """
        return sum(len(d['population']) for n, d in self.topology.nodes_iter(data=True))


    def __len__(self):
        """Return the length of a Metapopulation

        We'll define the length of a metapopulation as its size, so len(metapop)
        returns the number of individuals in all populations of Metapopulation
        metapop
        """
        return self.size()


    def num_producers(self):
        """Return the number of producers in the metapopulation"""
        return sum(d['population'].num_producers() for n, d in self.topology.nodes_iter(data=True))


    def prop_producers(self):
        """Get the proportion of producers in the metapopulation"""
        metapopsize = self.size()

        if metapopsize == 0:
            return 'NA'
        else:
            return 1.0 * self.num_producers() / self.size()


    def max_fitnesses(self):
        """Get the maximum fitness among producers and non-producers"""

        prod_max = [d['population'].max_fitnesses()[0] for n, d in self.topology.nodes_iter(data=True)]
        nonprod_max = [d['population'].max_fitnesses()[1] for n, d in self.topology.nodes_iter(data=True)]

        return (prod_max, nonprod_max)


    def write_logfiles(self):
        """Write any log files"""

        for (freq, l) in self.log_objects:
            if self.time % freq == 0:
                l.update(time=self.time)


    def cleanup(self):
        for (freq, l) in self.log_objects:
            l.close()

